# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab 2
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""Constants for the DHCP client implementation of the Anonymity Profile
([:rfc:`7844`])."""
import logging

logger = logging.getLogger(__name__)


DT_PRINT_FORMAT = '%y-%m-%d %H:%M:%S'

# client
##########
BROADCAST_MAC = 'ff:ff:ff:ff:ff:ff'
META_MAC = '00:00:00:00:00:00'
BROADCAST_ADDR = '255.255.255.255'
META_ADDR = '0.0.0.0'
CLIENT_PORT = 68
SERVER_PORT = 67

# DHCP timers
##########
LEASE_TIME = 1209600  # 14 DAYS
RENEWING_TIME = 604800  # 7 DAYS
REBINDING_TIME = 1058400  # 12 DAYS
DELAY_SELECTING = 10
TIMEOUT_SELECTING = 60
TIMEOUT_REQUESTING = 60
TIMEOUT_REQUEST_RENEWING = 226800
TIMEOUT_REQUEST_REBINDING = 75600

MAX_DELAY_SELECTING = 10
RENEW_PERC = 0.5
REBIND_PERC = 0.875

# DHCP number packet retransmissions
MAX_ATTEMPTS_DISCOVER = 5
MAX_OFFERS_COLLECTED = 1
MAX_ATTEMPTS_REQUEST = 5

# DHCP packet
##############
DHCP_OFFER_OPTIONS = [
    'server_id', 'subnet_mask', 'broadcast_address',
    'router', 'domain', 'name_server', 'lease_time', 'renewal_time',
    'rebinding_time']

# DHCP FSM
#############
STATE_ERROR = -1
STATE_PREINIT = 0
STATE_INIT = 1
STATE_SELECTING = 2
STATE_REQUESTING = 3
STATE_BOUND = 4
STATE_RENEWING = 5
STATE_REBINDING = 6
STATE_END = 7

STATES2NAMES = {
    STATE_ERROR: 'ERROR',
    STATE_PREINIT: 'PREINIT',
    STATE_INIT: 'INIT',
    STATE_SELECTING: 'SELECTING',
    STATE_REQUESTING: 'REQUESTING',
    STATE_BOUND: 'BOUND',
    STATE_RENEWING: 'RENEWING',
    STATE_REBINDING: 'REBINDING',
    STATE_END: 'END',
}

# NM integration
#####################
REASONS_NM = [
    'bound', 'renew', 'rebind',
    'timeout',
    'nak', 'expire',
    'end',
    'fail',
    'abend'
]

STATES2REASONS = {
    STATE_PREINIT: 'PREINIT',
    STATE_INIT: 'INIT',
    STATE_SELECTING: 'SELECTING',
    STATE_BOUND: 'BOUND',
    STATE_END: 'END',
    STATE_REBINDING: 'REBIND',
    STATE_RENEWING: 'RENEW',
    # "EXPIRE"
    STATE_ERROR: "TIMEOUT",
    # "FAIL"
    # "STOP"
}

# similar to systemd
DHCP_EVENTS = {
    'STOP': 0,
    'IP_ACQUIRE': 1,
    'IP_CHANGE': 2,
    'EXPIRED': 3,
    'RENEW': 4,
}

SCRIPT_ENV_KEYS = ['reason', 'medium', 'client', 'pid', 'interface',
                   'ip_address', 'subnet_mask', 'network_number',
                   'broadcast_address', 'domain_name_servers', 'routers',
                   'dhcp_server_identifier', 'next_server', 'domain_name',
                   'dhcp_lease_time', 'dhcp_renewal_time',
                   'dhcp_rebinding_time', 'expire', 'renew', 'rebind']

LEASEATTRS_SAMEAS_ENVKEYS = [
    'interface', 'subnet_mask', 'broadcast_address', 'next_server', 'rebind',
    'renew']

LEASEATTRS2ENVKEYS = {
    'ip_address': 'address',
    'routers': 'router',
    'network_number': 'subnet',
    'domain_name': 'domain',
    'domain_name_servers': 'name_server',
    'dhcp_lease_time': 'lease_time',
    'dhcp_rebinding_time': 'rebinding_time',
    'dhcp_renewal_time': 'renewal_time',
    'expire': 'expiry',
    'dhcp_server_identifier': 'server_id',
}

FSM_ATTRS = ['request_attempts', 'discover_attempts', 'script',
             'time_sent_request', 'current_state', 'client']



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""."""
import pytest
from dhcpcanon.dhcpcap import DHCPCAP


@pytest.fixture
def dhcpcap_maker(request):
    """ return a function which creates initialized dhcpcap instances. """

    def maker():
        dhcpcap = DHCPCAP(client_mac="00:01:02:03:04:05", iface='eth0')
        return dhcpcap
    return maker


@pytest.fixture
def dhcpcap(dhcpcap_maker):
    """ return an initialized dhcpcap instance. """
    return dhcpcap_maker()



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""DCHP client implementation of the Anonymity Profile [:rfc:`7844`]."""
from __future__ import absolute_import, unicode_literals

import logging

from netaddr import AddrFormatError
from scapy.arch import get_if_raw_hwaddr
from scapy.automaton import ATMT, Automaton
from scapy.config import conf
from scapy.utils import str2mac

from .clientscript import ClientScript
from .constants import (CLIENT_PORT, DELAY_SELECTING, FSM_ATTRS, LEASE_TIME,
                        MAX_ATTEMPTS_DISCOVER, MAX_ATTEMPTS_REQUEST,
                        MAX_OFFERS_COLLECTED, REBINDING_TIME, RENEWING_TIME,
                        SERVER_PORT, STATE_BOUND, STATE_END, STATE_ERROR,
                        STATE_INIT, STATE_PREINIT, STATE_REBINDING,
                        STATE_RENEWING, STATE_REQUESTING, STATE_SELECTING,
                        STATES2NAMES, TIMEOUT_REQUEST_REBINDING,
                        TIMEOUT_REQUEST_RENEWING, TIMEOUT_REQUESTING,
                        TIMEOUT_SELECTING)
from .dhcpcap import DHCPCAP
from .dhcpcaputils import isack, isnak, isoffer
from .timers import (gen_delay_selecting, gen_timeout_request_rebind,
                     gen_timeout_request_renew, gen_timeout_resend, nowutc)

logger = logging.getLogger(__name__)


class DHCPCAPFSM(Automaton):
    """DHCP client Finite State Machine (FSM).

    ... todo::

        - Group methods that do the same.
        - Check other implementations for the functionality not specifiyed in
          the RFCs.

    """

    def dict_self(self):
        """Return the self object attributes not inherited as dict."""
        return {k: v for k, v in self.__dict__.items() if k in FSM_ATTRS}

    def __str__(self):
        return str(self.dict_self())

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def reset(self, iface=None, client_mac=None):
        """Reset object attributes when state is INIT."""
        logger.debug('Reseting attributes.')
        if iface is None:
            iface = conf.iface
        if client_mac is None:
            _, mac = get_if_raw_hwaddr(iface)
            client_mac = str2mac(mac)
        self.client = DHCPCAP(iface, client_mac)
        self.script = ClientScript()
        self.time_sent_request = None
        self.discover_attempts = 0
        self.request_attempts = 0
        self.current_state = STATE_PREINIT
        self.offers = list()

    def __init__(self, iface=None, server_port=None,
                 client_port=None, client_mac=None,
                 scriptfile=None, delay_selecting=None, timeout_select=None,
                 debug_level=5, *args, **kargs):
        """Overwrites Automaton __init__ method.

        [ :rfc:`7844#3.4` ] ::
            If the hardware address is reset to a new
            randomized value, the DHCP client SHOULD use the new randomized
            value in the DHCP messages

        """
        logger.debug('Inizializating FSM.')
        super(DHCPCAPFSM, self).__init__(*args, **kargs)
        self.debug_level = debug_level
        self.delay_selecting = delay_selecting
        self.timeout_select = timeout_select
        self.reset(iface, client_mac)
        self.client.server_port = server_port or SERVER_PORT
        self.client.client_port = client_port or CLIENT_PORT
        self.socket_kargs = {
            'filter': 'udp and src port {0} and dst port {1}'
                      ' and ether dst {2}'.
                      format(self.client.server_port,
                             self.client.client_port,
                             self.client.client_mac)
        }
        self.script.scriptname = scriptfile
        self.script.script_init(self.client.lease, self.current_state)
        self.script.script_go()
        logger.debug('FSM thread id: %s.', self.threadid)

    def get_timeout(self, state, function):
        """Workaround to get timeout in the ATMT.timeout class method."""
        state = STATES2NAMES[state]
        for timeout_fn_t in self.timeout[state]:
            # access the function name
            if timeout_fn_t[1] is not None and \
               timeout_fn_t[1].atmt_condname == function.atmt_condname:
                logger.debug('Timeout for state %s, function %s, is %s',
                             state, function.atmt_condname, timeout_fn_t[0])
                return timeout_fn_t[0]
        return None

    def set_timeout(self, state, function, newtimeout):
        """
        Workaround to change timeout values in the ATMT.timeout class method.

        self.timeout format is::

            {'STATE': [
                (TIMEOUT0, <function foo>),
                (TIMEOUT1, <function bar>)),
                (None, None)
                ],
            }


        """
        state = STATES2NAMES[state]
        for timeout_fn_t in self.timeout[state]:
            # access the function name
            if timeout_fn_t[1] is not None and \
               timeout_fn_t[1].atmt_condname == function.atmt_condname:
                # convert list to tuple to make it mutable
                timeout_l = list(timeout_fn_t)
                # modify the timeout
                timeout_l[0] = newtimeout
                # set the new timeoute to self.timeout
                i = self.timeout[state].index(timeout_fn_t)
                self.timeout[state][i] = tuple(timeout_l)
                logger.debug('Set state %s, function %s, to timeout %s',
                             state, function.atmt_condname, newtimeout)

    def send_discover(self):
        """Send discover."""
        assert self.client
        assert self.current_state == STATE_INIT or \
            self.current_state == STATE_SELECTING
        pkt = self.client.gen_discover()
        conf.checkIPaddr = 0
        self.my_send(pkt)
        conf.checkIPaddr = 1
        # FIXME: check that this is correct,: all or only discover?
        if self.discover_attempts < MAX_ATTEMPTS_DISCOVER:
            self.discover_attempts += 1
        timeout = gen_timeout_resend(self.discover_attempts)
        self.set_timeout(self.current_state,
                         self.timeout_selecting,
                         timeout)
        # logger.info('DHCPDISCOVER on %s to %s port %s' %
        #             (self.client.iface, self.client.server_mac,
        #              self.client.server_port)))

    def select_offer(self):
        """Select an offer from the offers received.

        [:rfc:`2131#4.2`]::

            DHCP clients are free to use any strategy in selecting a DHCP
            server among those from which the client receives a DHCPOFFER.

        [:rfc:`2131#4.4.1`]::

            The time
            over which the client collects messages and the mechanism used to
            select one DHCPOFFER are implementation dependent.

        Nor [:rfc:`7844`] nor [:rfc:`2131`] specify the algorithm.
        Here, currently the first offer is selected.

        .. todo::
             * Check other implementations algorithm to select offer.
        """
        logger.debug('Selecting offer.')
        pkt = self.offers[0]
        self.client.handle_offer(pkt)

    def send_request(self):
        """Send request."""
        assert self.client
        if self.current_state == STATE_BOUND:
            pkt = self.client.gen_request_unicast()
        else:
            pkt = self.client.gen_request()
            conf.checkIPaddr = 0
        self.my_send(pkt)
        conf.checkIPaddr = 1
        logger.debug('Modifying FSM obj, setting time_sent_request.')
        self.time_sent_request = nowutc()
        logger.info('DHCPREQUEST of %s on %s to %s port %s',
                    self.client.iface, self.client.client_ip,
                    self.client.server_ip, self.client.server_port)

        # FIXME: check that this is correct,: all of only discover?
        # and if > MAX_DISCOVER_RETRIES?
        if self.request_attempts < MAX_ATTEMPTS_REQUEST:
            self.request_attempts *= 2
            logger.debug('Increased request attempts to %s',
                         self.request_attempts)
        if self.current_state == STATE_RENEWING:
            timeout_renewing = gen_timeout_request_renew(self.client.lease)
            self.set_timeout(self.current_state,
                             self.timeout_request_renewing,
                             timeout_renewing)
        elif self.current_state == STATE_REBINDING:
            timeout_rebinding = gen_timeout_request_rebind(self.client.lease)
            self.set_timeout(self.current_state,
                             self.timeout_request_rebinding,
                             timeout_rebinding)
        else:
            timeout_requesting = \
                gen_timeout_resend(self.request_attempts)
            self.set_timeout(self.current_state,
                             self.timeout_requesting,
                             timeout_requesting)

    def set_timers(self):
        """Set renewal, rebinding times."""
        logger.debug('setting timeouts')
        self.set_timeout(self.current_state,
                         self.renewing_time_expires,
                         self.client.lease.renewal_time)
        self.set_timeout(self.current_state,
                         self.rebinding_time_expires,
                         self.client.lease.rebinding_time)

    def process_received_ack(self, pkt):
        """Process a received ACK packet.

        Not specifiyed in[:rfc:`7844`], [:rfc:`2131#2.2.`]::

            the allocating
            server SHOULD probe the reused address before allocating the
            address, e.g., with an ICMP echo request, and the client SHOULD
            probe the newly received address, e.g., with ARP.

        """
        if isack(pkt):
            # FIXME: check the fields match the previously offered ones?
            try:
                self.event = self.client.handle_ack(pkt,
                                                    self.time_sent_request)
            except AddrFormatError as err:
                logger.error(err)
                # the net values are not valid, go back to SELECTING state
                # (or just previous state?)
                raise self.SELECTING()
            # TODO: if address is taken (PING?) go to INIT and send DHCPDELINE
            logger.info('DHCPACK of %s from %s' %
                        (self.client.client_ip, self.client.server_ip))
            return True
        return False

    def process_received_nak(self, pkt):
        """Process a received NAK packet."""
        if isnak(pkt):
            logger.info('DHCPNAK of %s from %s',
                        self.client.client_ip, self.client.server_ip)
            return True
        return False

    #################################################################
    # State machine
    #################################################################
    @ATMT.state(initial=1)
    def INIT(self):
        """INIT state."""
        # in case INIT is reached from other state, initialize attributes
        # reset all variables.
        logger.debug('INIT')
        if self.current_state is not STATE_PREINIT:
            self.reset()
        self.current_state = STATE_INIT
        # [:rfc:`2131#4.4.1`]::
        # The client SHOULD wait a random time between one and ten
        #  seconds to desynchronize the use of DHCP at startup
        if self.delay_selecting is None:
            delay_selecting = gen_delay_selecting()
        else:
            delay_selecting = self.delay_selecting
        self.set_timeout(self.current_state,
                         self.timeout_delay_selecting,
                         delay_selecting)
        if self.timeout_select is not None:
            self.set_timeout(STATE_SELECTING,
                             self.timeout_selecting,
                             self.timeout_select)

    @ATMT.timeout(INIT, DELAY_SELECTING)
    def timeout_delay_selecting(self):
        """Timeout of delay selecting on INIT state."""
        raise self.SELECTING()

    @ATMT.action(timeout_delay_selecting)
    def on_timeout_delay_selecting(self):
        """Action on timeout of delay selecting on INIT state."""
        self.send_discover()
        logger.debug('DISCOVER sent')

    @ATMT.state()
    def SELECTING(self):
        """SELECTING state."""
        logger.debug('SELECTING')
        self.current_state = STATE_SELECTING

    @ATMT.receive_condition(SELECTING)
    def receive_offer(self, pkt):
        """Receive offer on SELECTING state."""
        if isoffer(pkt):
            logger.debug('OFFER received')
            self.offers.append(pkt)
            if len(self.offers) >= MAX_OFFERS_COLLECTED:
                self.select_offer()
                raise self.REQUESTING()
            else:
                # FIXME: neeeded?
                raise self.SELECTING()

    @ATMT.action(receive_offer)
    def on_select_offer(self):
        """Action on receive offer on SELECTING state."""
        self.send_request()

    @ATMT.timeout(SELECTING, TIMEOUT_SELECTING)
    def timeout_selecting(self):
        """Timeout of selecting on SELECTING state.

        Not specifiyed in [:rfc:`7844#`].See comments in
        :func:`dhcpcapfsm.DHCPCAPFSM.timeout_request`.

        """
        if self.discover_attempts >= MAX_ATTEMPTS_DISCOVER:
            logger.debug('Maximum number of discover retries is %s'
                         ' and already sent %s',
                         MAX_ATTEMPTS_DISCOVER, self.discover_attempts)
            if len(self.offers) < 1:
                logger.debug('No offer was received')
                raise self.ERROR()
            else:
                # FIXME: correct?
                logger.debug('needed? Use the offers received')
                raise self.REQUESTING()
        # else self.discover_attempts < MAX_ATTEMPTS_DISCOVER
        if len(self.offers) < MAX_OFFERS_COLLECTED:
            logger.debug('needed? Naximum number of offers not reached')
            raise self.SELECTING()

        logger.debug('Naximum number of offers reached')
        raise self.REQUESTING()

    @ATMT.action(timeout_selecting)
    def on_retransmit_discover(self):
        """Action on retransmit discover on SELECTING state."""
        self.send_discover()
        logger.debug('DISCOVER resent')

    @ATMT.state(error=1)
    def ERROR(self):
        """ERROR state."""
        logger.debug('ERROR')
        self.current_state = STATE_ERROR
        self.script.script_init(self.client.lease, self.current_state)
        self.script.script_go()
        raise self.END()

    @ATMT.state()
    def REQUESTING(self):
        """REQUESTING state."""
        logger.debug('REQUESTING')
        self.current_state = STATE_REQUESTING

    @ATMT.receive_condition(REQUESTING)
    def receive_ack_requesting(self, pkt):
        """Receive ack on REQUESTING state."""
        if self.process_received_ack(pkt):
            raise self.BOUND()

    @ATMT.receive_condition(REQUESTING)
    def receive_nak_requesting(self, pkt):
        """Receive nak on REQUESTING state."""
        if self.process_received_nak(pkt):
            raise self.INIT()

    @ATMT.action(receive_ack_requesting)
    def on_ack_requesting(self):
        """Action on ack requesting on REQUESTING state."""
        # [:rfc:`7844`]: not recording lease
        logger.debug('Setting timers.')
        self.set_timers()

    @ATMT.timeout(REQUESTING, TIMEOUT_REQUESTING)
    def timeout_requesting(self):
        """Timeout of requesting on REQUESTING state.

        Not specifiyed in [:rfc:`7844`]

        [:rfc:`2131#3.1`]::

            might retransmit the
            DHCPREQUEST message four times, for a total delay of 60 seconds

        """
        if self.discover_requests >= MAX_ATTEMPTS_REQUEST:
            logger.debug('Maximum number of reuqest retries reached'
                         ' is %s and already sent %s',
                         MAX_ATTEMPTS_REQUEST, self.disover_requests)
            raise self.ERROR()
        raise self.REQUESTING()

    @ATMT.action(timeout_requesting)
    def on_retransmit_request(self):
        """Action on timeout of requesting on REQUESTING state.

        Send REQUEST.

        """
        self.send_request()

    @ATMT.state()
    def BOUND(self):
        """BOUND state."""
        logger.debug('BOUND')
        logger.info('(%s) state changed %s -> bound', self.client.iface,
                    STATES2NAMES[self.current_state])
        self.current_state = STATE_BOUND
        self.script.script_init(self.client.lease, self.current_state)
        self.script.script_go()
        # TODO: go daemon?

    @ATMT.timeout(BOUND, RENEWING_TIME)
    def renewing_time_expires(self):
        """Timout of renewing time."""
        raise self.RENEWING()

    @ATMT.action(renewing_time_expires)
    def on_renewing_time_expires(self):
        """Action on renewing time expires on BOUND state."""
        # FIXME: udp
        self.send_request()

    @ATMT.state()
    def RENEWING(self):
        """RENEWING state."""
        self.current_state = STATE_RENEWING
        self.script.script_init(self.client.lease, self.current_state)
        self.script.script_go()

    @ATMT.receive_condition(RENEWING)
    def receive_ack_renewing(self, pkt):
        """Receive ack on RENEWING state."""
        if self.process_received_ack(pkt):
            raise self.BOUND()

    @ATMT.receive_condition(RENEWING)
    def receive_nak_renewing(self, pkt):
        """Receive nak on RENEWING state."""
        if self.process_received_nak(pkt):
            raise self.INIT()

    @ATMT.action(receive_ack_renewing)
    def on_renewing(self):
        """Action on renewing on RENEWING state.

        Not recording lease, but restarting timers.

        """
        self.client.lease.sanitize_net_values()
        self.client.lease.set_times(self.time_sent_request)
        self.set_timers()

    @ATMT.timeout(RENEWING, TIMEOUT_REQUEST_RENEWING)
    def timeout_request_renewing(self):
        """Timeout of renewing on RENEWING state.

        Same comments as in
        :func:`dhcpcapfsm.DHCPCAPFSM.timeout_requesting`.

        """
        if self.request_attempts >= MAX_ATTEMPTS_REQUEST:
            logger.debug('Maximum number of request retries renewing is'
                         ' %s, already sent %s.',
                         MAX_ATTEMPTS_REQUEST, self.request_attempts)
            raise self.ERROR()
        raise self.RENEWING()

    @ATMT.action(timeout_request_renewing)
    def on_retransmit_request_renewing(self):
        """Action on timeout of request newing on RENEWING state.

        Send REQUEST.

        """
        self.send_request()

    @ATMT.timeout(RENEWING, REBINDING_TIME)
    def rebinding_time_expires(self):
        """Timeout of rebinding time on RENEWING state."""
        raise self.REBINDING()

    @ATMT.action(rebinding_time_expires)
    def on_rebinding_time_expires(self):
        """Action on rebinding time expires on RENEWING state."""
        self.send_request()

    @ATMT.state()
    def REBINDING(self):
        """REBINDING state."""
        self.current_state = STATE_REBINDING
        self.script.script_init(self.client.lease, self.current_state)
        self.script.script_go()

    @ATMT.timeout(REBINDING, LEASE_TIME)
    def lease_expires(self):
        """Timeout of lease on REBINDING state.
        Not sending DHCPRELEASE to minimize deanonymization

        [:rfc:`2131#4.4.6`]::

            Note that the correct operation
            of DHCP does not depend on the transmission of DHCPRELEASE.

        """
        raise self.STATE_INIT()

    @ATMT.receive_condition(REBINDING)
    def receive_ack_rebinding(self, pkt):
        """Receive ack on REBINDING state."""
        if self.process_received_ack(pkt):
            raise self.BOUND()

    @ATMT.receive_condition(REBINDING)
    def receive_nak_rebinding(self, pkt):
        """Receive nak on REBINDING state."""
        if self.process_received_nak(pkt):
            raise self.INIT()

    @ATMT.action(receive_ack_rebinding)
    def on_rebinding(self):
        """Action on receive ACK rebinding on REBINDING state.

        Not recording lease, but start new lease

        """
        self.set_timers()

    @ATMT.timeout(REBINDING, TIMEOUT_REQUEST_REBINDING)
    def timeout_request_rebinding(self):
        """Timeout of request rebinding on REBINDING state.

        Same comments as in
        :func:`dhcpcapfsm.DHCPCAPFSM.timeout_requesting`.

        """
        if self.request_attempts >= MAX_ATTEMPTS_REQUEST:
            logger.debug('Maximum number of request retries rebinding is'
                         ' %s, already sent %s.',
                         MAX_ATTEMPTS_REQUEST, self.request_attempts)
            raise self.ERROR()
        raise self.REBINDING()

    @ATMT.action(timeout_request_rebinding)
    def on_retransmit_request_rebinding(self):
        """Action on request rebinding on REBINDING state.

        Send REQUEST.

        """
        self.send_request()

    @ATMT.state(final=1)
    def END(self):
        """END state."""
        logger.debug('END')
        self.current_state = STATE_END
        self.script.script_init(self.client.lease, self.current_state)
        self.script.script_go()
        self.reset()



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab 2
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""__init__ for the DHCP client implementation of the Anonymity Profile
([:rfc:`7844`])."""
from __future__ import absolute_import
# from . import (clientscript, conflog, constants, dhcpcap, dhcpcapfsm,
#                dhcpcaplease, dhcpcaputils, timers)
try:
    from ._version import version
except ImportError:
    try:
        from setuptools_scm import get_version
        version = get_version()
    except (ImportError, LookupError):
        version = '0.3.0'

__version__ = version
__author__ = "juga"
__author_mail__ = "juga@riseup.net"
__description__ = "DHCP client disclosing less identifying information"
__long_description__ = "Python implmentation of the DHCP Anonymity Profile \
                        (RFC7844) designed for users that \
                        wish to remain anonymous to the visited network \
                        minimizing disclosure of identifying information."
__website__ = 'https://github.com/juga0/dhcpcanon'
__documentation__ = 'http://dhcpcanon.readthedocs.io/en/' + __version__
__authors__ = []
__copyright__ = """Copyright (C) 2016 <juga@riseup.net>
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions.
For details see the COPYRIGHT file distributed along this program."""

__license__ = """
    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    This package is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this package. If not, see <http://www.gnu.org/licenses/>.
"""
__all__ = ('clientscript', 'conflog', 'dhcpcapfsm', 'dhcpcaplease',
           'dhcpcaputils', 'timers', 'constants', 'dhcpcap')



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab 2
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""Util functions for the DHCP client implementation of the Anonymity Profile
([:rfc:`7844`])."""
from __future__ import absolute_import

import logging

from scapy.arch.linux import get_if_list
from scapy.layers.dhcp import DHCP, DHCPTypes

logger = logging.getLogger(__name__)


def isoffer(packet):
    """."""
    if DHCP in packet and (DHCPTypes.get(packet[DHCP].options[0][1]) == 'offer'
                           or packet[DHCP].options[0][1] == "offer"):
        logger.debug('Packet is Offer.')
        return True
    return False


def isnak(packet):
    """."""
    if DHCP in packet and (DHCPTypes.get(packet[DHCP].options[0][1]) == 'nak'
                           or packet[DHCP].options[0][1] == 'nak'):
        logger.debug('Packet is NAK.')
        return True
    return False


def isack(packet):
    """."""
    if DHCP in packet and (DHCPTypes.get(packet[DHCP].options[0][1]) == 'ack'
                           or packet[DHCP].options[0][1] == 'ack'):
        logger.debug('Packet is ACK.')
        return True
    return False


def discover_ifaces():
    ifaces = get_if_list()
    ifaces.remove('lo')
    logger.debug('Disovered interfaces %s.', ifaces)
    return ifaces


def detect_speed_network():
    # 100 Mbps = 100 Mb/s
    with open('/sys/class/net/eth0/speed') as fd:
        speed = fd.read()
    logger.debug('Net speed %s', speed)
    return speed


# TODO
def detect_initial_network():
    pass



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab 2
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""Client class for the DHCP client implementation of the Anonymity Profile
([:rfc:`7844`])."""
from __future__ import absolute_import

# from __future__ import unicode_literals
import logging

import attr
from netaddr import IPNetwork
from scapy.arch import get_if_raw_hwaddr
from scapy.config import conf
from scapy.layers.dhcp import BOOTP, DHCP
from scapy.layers.inet import IP, UDP
from scapy.layers.l2 import Ether
from scapy.utils import mac2str, str2mac

from .constants import (BROADCAST_ADDR, BROADCAST_MAC, CLIENT_PORT,
                        DHCP_EVENTS, DHCP_OFFER_OPTIONS, META_ADDR,
                        SERVER_PORT)
from .dhcpcaplease import DHCPCAPLease

logger = logging.getLogger('dhcpcanon')


@attr.s
class DHCPCAP(object):
    """."""
    iface = attr.ib(default=None)

    client_mac = attr.ib(default=None)
    client_ip = attr.ib(default=META_ADDR)
    client_port = attr.ib(default=CLIENT_PORT)

    server_mac = attr.ib(default=BROADCAST_MAC)
    server_ip = attr.ib(default=BROADCAST_ADDR)
    server_port = attr.ib(default=SERVER_PORT)

    lease = attr.ib(default=attr.Factory(DHCPCAPLease))
    event = attr.ib(default=None)

    def __attrs_post_init__(self):
        """Initializes attributes after attrs __init__.

        These attributes do not change during the life of the object.

        """
        if self.iface is None:
            self.iface = conf.iface
        if self.client_mac is None:
            _, client_mac = get_if_raw_hwaddr(self.iface)
            self.client_mac = str2mac(client_mac)
        logger.debug('Modifying Lease obj, setting iface.')
        self.lease.interface = self.iface

    def gen_ether_ip(self):
        """Generates link layer and IP layer part of DHCP packet.

        For broadcast packets is:
            Ether(src=client_mac, dst="ff:ff:ff:ff:ff:ff") /
            IP(src="0.0.0.0", dst="255.255.255.255") /

        """
        ether_ip = (Ether(src=self.client_mac, dst=BROADCAST_MAC) /
                    IP(src=META_ADDR, dst=BROADCAST_ADDR))
        return ether_ip

    def gen_ether_ip_unicast(self):
        """Generates link layer and IP layer part of DHCP packet.

        For unicast packets is:
            Ether(src=client_mac, dst=server_mac) /
            IP(src=client_ip?, dst=server_ip) /

        """
        ether_ip = (Ether(src=self.client_mac, dst=self.server_mac) /
                    IP(src=self.client_ip, dst=self.server_ip))
        return ether_ip

    def gen_udp(self):
        """Generates UDP layer part of DHCP packet.

        UDP layer is always:
            UDP(sport=68, dport=67) /

        """
        udp = (UDP(sport=self.client_port, dport=self.server_port))
        return udp

    def gen_bootp(self):
        """Generates BOOTP layer part of DHCP packet.

        [ :rfc:`7844#3.4` ] ::

            The presence of this address is necessary for the proper operation
            of the DHCP service.

        [:rfc:`7844#3.`] ::
            MAY contain the Client Identifier option,

        """
        bootp = (
            BOOTP(chaddr=[mac2str(self.client_mac)])
            # , ciaddr=META_ADDR)
        )
        return bootp

    def gen_bootp_unicast(self):
        """Generates BOOTP layer part of unicast DHCP packet.

        Same comments as in gen_bootp

        """
        bootp = (
            BOOTP(chaddr=[mac2str(self.client_mac)])
            # , ciaddr=self.client_ip)
        )
        return bootp

    def gen_discover(self):
        """
        Generate DHCP DISCOVER packet.

        [:rfc:`7844#3.1`] ::

            SHOULD randomize the ordering of options

            If this can not be implemented
            MAY order the options by option code number (lowest to highest).

        [:rfc:`7844#3.`] ::
            MAY contain the Parameter Request List option.

        """
        dhcp_discover = (
            self.gen_ether_ip() /
            self.gen_udp() /
            self.gen_bootp() /
            DHCP(options=[
                ("message-type", "discover"),
                "end"
            ])
        )
        logger.debug('Generated discover %s.', dhcp_discover.summary())
        return dhcp_discover

    def gen_request(self):
        """
        Generate DHCP REQUEST packet.

        [:rfc:`7844#3.1`] ::

            SHOULD randomize the ordering of options

            If this can not be implemented
            MAY order the options by option code number (lowest to highest).

        [:rfc:`7844#3.`] ::
            MAY contain the Parameter Request List option.

        If in response to a DHCPOFFER,::

            MUST contain the corresponding Server Identifier option
            MUST contain the Requested IP address option.

            If the message is not in response to a DHCPOFFER (BOUND, RENEW),::
            MAY contain a Requested IP address option

        """
        dhcp_req = (
            self.gen_ether_ip() /
            self.gen_udp() /
            self.gen_bootp() /
            DHCP(options=[
                ("message-type", "request"),
                ("requested_addr", self.lease.address),
                ("server_id", self.lease.server_id),
                "end"])
        )
        logger.debug('Generated request %s.', dhcp_req.summary())
        return dhcp_req

    def gen_request_unicast(self):
        """
        Generate DHCP REQUEST unicast packet.

        Same comments as in gen_request apply.

        """
        dhcp_req = (
            self.gen_ether_ip_unicast() /
            self.gen_udp() /
            self.gen_bootp_unicast() /
            DHCP(options=[
                ("message-type", "request"),
                "end"])
        )
        logger.debug('Generated request %s.', dhcp_req.summary())
        return dhcp_req

    def gen_decline(self):
        """
        Generate DHCP decline packet (broadcast).

        [:rfc:`7844#3.`] ::

            MUST contain the Message Type option,
            MUST contain the Server Identifier option,
            MUST contain the Requested IP address option;

        .. note:: currently not being used.

        """
        dhcp_decline = (
            self.gen_ether_ip() /
            self.gen_udp() /
            self.gen_bootp() /
            DHCP(options=[
                ("message-type", "decline"),
                ("server_id", self.server_ip),
                ("requested_addr", self.client_ip),
                "end"])
        )
        logger.debug('Generated decline.')
        logger.debug(dhcp_decline.summary())
        return dhcp_decline

    def gen_release(self):
        """
        Generate DHCP release packet (broadcast?).

        [:rfc:`7844#3.`] ::

            MUST contain the Message Type option and
            MUST contain the Server Identifier option,

        .. note:: currently not being used.

        """
        dhcp_release = (
            self.gen_ether_ip() /
            self.gen_udp() /
            self.gen_bootp() /
            DHCP(options=[
                ("message-type", "release"),
                ("server_id", self.server_ip),
                "end"])
        )
        logger.debug('Generated release.')
        logger.debug(dhcp_release.summary())
        return dhcp_release

    def gen_inform(self):
        """
        Generate DHCP inform packet (unicast).

        [:rfc:`7844#3.`] ::

            MUST contain the Message Type option,

        .. note:: currently not being used.

        """
        dhcp_inform = (
            self.gen_ether_ip_unicast() /
            self.gen_udp() /
            self.gen_bootp_unicast() /
            DHCP(options=[
                ("message-type", "inform"),
                "end"])
        )
        logger.debug('Generated inform.')
        logger.debug(dhcp_inform.summary())
        return dhcp_inform

    def gen_net_values(self, addr, netmask, router):
        """Generate network mask in CIDR format and subnet.

        Validate the given arguments. Otherwise AddrFormatError exception
        will be raised and catched in the FSM.

        """
        ipn = IPNetwork(addr + '/' + netmask)
        ripn = IPNetwork(router + '/' + netmask)
        assert ripn.network == ipn.network
        logger.debug('Net values are valid')
        return {'subnet_mask_cidr': str(ipn.prefixlen),
                'subnet': str(ipn.network)}

    def handle_offer_ack(self, pkt, time_sent_request=None):
        """Create a lease object with the values in OFFER/ACK packet."""
        attrs_dict = dict([(opt[0], str(opt[1])) for opt in pkt[DHCP].options
                           if isinstance(opt, tuple)
                           and opt[0] in DHCP_OFFER_OPTIONS])
        net_attrs_dict = self.gen_net_values(pkt[BOOTP].yiaddr,
                                             attrs_dict.get('subnet_mask'),
                                             attrs_dict.get('router'))
        attrs_dict.update(net_attrs_dict)
        attrs_dict.update({
            "interface": self.iface,
            "address": pkt[BOOTP].yiaddr,
            "next_server": pkt[BOOTP].siaddr,
        })
        logger.debug('Creating Lease obj.')
        lease = DHCPCAPLease(**attrs_dict)
        return lease

    def handle_offer(self, pkt):
        """."""
        logger.debug("Handling Offer.")
        logger.debug('Modifying obj DHCPCAP, setting lease.')
        self.lease = self.handle_offer_ack(pkt)

    def handle_ack(self, pkt, time_sent_request):
        """."""
        logger.debug("Handling ACK.")
        logger.debug('Modifying obj DHCPCAP, setting server data.')
        self.server_mac = pkt[Ether].src
        self.server_ip = pkt[IP].src
        self.server_port = pkt[UDP].sport
        event = DHCP_EVENTS['IP_ACQUIRE']
        # FIXME: check the fields match the previously offered ones?
        # FIXME: create a new object also on renewing/rebinding
        # or only set_times?
        lease = self.handle_offer_ack(pkt, time_sent_request)
        lease.set_times(time_sent_request)
        if self.lease is not None:
            if (self.lease.address != lease.address or
                    self.lease.subnet_mask != lease.subnet_mask or
                    self.lease.router != lease.router):
                event = DHCP_EVENTS['IP_CHANGE']
            else:
                event = DHCP_EVENTS['RENEW']
        logger.debug('Modifying obj DHCPCAP, setting lease, client ip, event.')
        self.lease = lease
        self.client_ip = self.lease.address
        self.event = event
        return event



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.

"""Setup."""
from setuptools import find_packages, setup

import dhcpcanon

setup(
    name='dhcpcanon',
    version=dhcpcanon.__version__,
    description=dhcpcanon.__description__,
    long_description=dhcpcanon.__long_description__,
    author=dhcpcanon.__author__,
    author_email=dhcpcanon.__author_mail__,
    license='MIT',
    url=dhcpcanon.__website__,
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=[
        'scapy>=2.2";python_version<="2.7"',
        'scapy-python3>=0.21;python_version>="3.4"',
        "netaddr>=0.7",
        "pip>=8.1",
        "attrs>=16.3",
        "daemon>=1.1"
    ],
    # leaved commented to have concrete dependencies
    # dependency_links=[
    #     "https://pypi.python.org/simple/scapy==2.2.0-dev",
    #     "https://pypi.python.org/simple/netaddr==0.7.10",
    #     "https://pypi.python.org/simple/ipaddr==2.1.11",
    #     "https://pypi.python.org/simple/pytz==2016.6.1",
    #     "https://pypi.python.org/simple/pip==8.1.2",
    #     "https://pypi.python.org/simple/pyroute2==0.4.11"
    #     "https://pypi.python.org/simple/attrs==16.3.0"
    # ],
    extras_require={
        'dev': ['ipython', 'pyflakes', 'pep8'],
        'test': ['coverage', 'coveralls', 'codecov', 'tox', 'pytest'],
        'doc': ['sphinx']
    },
    # entry_points={
    #     'console_scripts': [
    #         'dhcpcanon = scripts.dhcpcanon:main',
    #     ]
    # },
    scripts=['scripts/dhcpcanon'],
    keywords='python scapy dhcp RFC7844 RFC2131 anonymity',
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Environment :: Console",
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        "Topic :: System :: Networking",
    ],
)



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab 2
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""Class to Initialize and call external script."""
from __future__ import absolute_import, unicode_literals

import logging
import os
import subprocess

import attr

from .constants import (LEASEATTRS2ENVKEYS, LEASEATTRS_SAMEAS_ENVKEYS,
                        SCRIPT_ENV_KEYS, STATES2REASONS)

logger = logging.getLogger('dhcpcanon')


@attr.s
class ClientScript(object):
    """Simulates the behaviour of the isc-dhcp client-script or nm-dhcp-helper.

    `client-script
    <https://anonscm.debian.org/cgit/pkg-dhcp/isc-dhcp.git/tree/client/scripts/linux>`_
    or `nm-dhcp-helper
    <https://github.com/NetworkManager/NetworkManager/tree/master/src/dhcp>`_.

    """

    scriptname = attr.ib(default=None)
    env = attr.ib(default=attr.Factory(dict))

    def __attrs_post_init__(self, env=None):
        """."""
        logger.debug('Modifying ClientScript obj after creating it.')
        if env is None:
            self.env = dict.fromkeys(SCRIPT_ENV_KEYS, str(''))
        else:
            self.env = env

    def script_init(self, lease, state, prefix='', medium=''):
        """Initialize environment to pass to the external script."""
        if self.scriptname is not None:
            logger.debug('Modifying ClientScript obj, setting env.')
            if isinstance(state, int):
                reason = STATES2REASONS[state]
            else:
                reason = state
            self.env['reason'] = str(reason)
            self.env['medium'] = str(medium)
            self.env['client'] = str('dhcpcanon')
            self.env['pid'] = str(os.getpid())

            for k in LEASEATTRS_SAMEAS_ENVKEYS:
                self.env[k] = str(lease.__getattribute__(k))

            for k, v in LEASEATTRS2ENVKEYS.items():
                self.env[k] = str(lease.__getattribute__(v))
        else:
            logger.debug('There is not script path.')

    def script_go(self, scriptname=None, env=None):
        """Run the external script."""
        scriptname = self.scriptname or scriptname
        if scriptname is not None:
            env = self.env or env
            logger.debug('Calling script %s', scriptname)
            sp = subprocess.Popen([scriptname], stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE, env=env,
                                  close_fds=True, shell=False)
            _, err = sp.communicate()
            if sp.returncode != 0:
                logger.debug('sp err %s', err)
            return sp.returncode
        return None



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab 2
# Copyright 2016, 2017 juga (juga at riseup dot net), MIT license.
"""Lease class for the DHCP client implementation of the Anonymity Profile
([:rfc:`7844`]).."""
from __future__ import absolute_import

import logging

import attr
from attr.validators import instance_of

from .timers import (future_dt_str, gen_rebinding_time, gen_renewing_time,
                     nowutc)

logger = logging.getLogger('dhcpcanon')


@attr.s
class DHCPCAPLease(object):
    """."""
    address = attr.ib(default='', validator=instance_of(str))
    server_id = attr.ib(default='', validator=instance_of(str))
    next_server = attr.ib(default='', validator=instance_of(str))
    router = attr.ib(default='', validator=instance_of(str))
    subnet_mask = attr.ib(default='', validator=instance_of(str))
    broadcast_address = attr.ib(default='', validator=instance_of(str))
    domain = attr.ib(default='', validator=instance_of(str))
    name_server = attr.ib(default='', validator=instance_of(str))
    subnet = attr.ib(default='', validator=instance_of(str))
    lease_time = attr.ib(default='', validator=instance_of(str))
    renewal_time = attr.ib(default='', validator=instance_of(str))
    rebinding_time = attr.ib(default='', validator=instance_of(str))
    # not given by the server
    interface = attr.ib(default='', validator=instance_of(str))
    # not given by the server, calculated on previous
    subnet_mask_cidr = attr.ib(default='', validator=instance_of(str))
    network = attr.ib(default='', validator=instance_of(str))
    expiry = attr.ib(default='', validator=instance_of(str))
    renew = attr.ib(default='', validator=instance_of(str))
    rebind = attr.ib(default='', validator=instance_of(str))

    # def __attrs_post_init__(self, sent_dt):
    #     """Initializes attributes after attrs __init__."""
    #     self.set_times(sent_dt)

    def set_times(self, sent_dt):
        """
        Set timers for the lease given the time in which the request was sent.

        [:rfc:`2131#4.4.1`]::

            The client records the lease expiration time
            as the sum of the time at which the original request was
            sent and the duration of the lease from the DHCPACK message.

        """
        logger.debug('Modifying Lease obj, setting timers.')
        elapsed = (nowutc() - sent_dt).seconds
        if self.renewal_time == '':
            self.renewal_time = gen_renewing_time(self.lease_time, elapsed)
        if self.rebinding_time == '':
            self.rebinding_time = gen_rebinding_time(self.lease_time, elapsed)
        self.expiry = future_dt_str(sent_dt, self.lease_time)
        self.renew = future_dt_str(sent_dt, self.renewal_time)
        self.rebind = future_dt_str(sent_dt, self.rebinding_time)
        logger.debug('lease time: %s, expires on %s', self.lease_time,
                     self.expiry)
        logger.debug('renewal_time: %s, expires on %s',
                     self.renewal_time, self.renew)
        logger.debug('rebinding time: %s, expires on %s',
                     self.rebinding_time, self.rebind)

    def info_lease(self):
        """Print lease information."""
        logger.info('address %s', self.address)
        logger.info('plen %s (%s)', self.subnet_mask_cidr, self.subnet_mask)
        logger.info('gateway %s', self.router)
        logger.info('server identifier %s', self.server_id)
        logger.info('nameserver %s', self.name_server)
        logger.info('domain name %s', self.domain)
        logger.info('lease time %s', self.lease_time)
